%**********************************************************
%
%   suemenu.sl defines and creates menubar and pulldowns
%   upon call from sue.sl.
%   It includes some functions from the Help pulldown
%   which are only used here.
%
%**********************************************************

    _traceback = 1;


%---    required prototypes  ------------------------------

define guide_keymap ()    ;
define keyhelp__keymap () ;
define prghelp_clip ()    ;
define prghelp_keymap ()  ;
define prghelp__keymap () ;
define viewer_load()      ;

%----------------------------------------------------------


variable explainkey = String_Type [116,2];
%   first  column: functions that have a key shortcut
%   second column: explanatory comments on them


define display_help (stringA, stringB, stringC)
%----------------------------------------------------------
%   displays selected text of help file
%   relying on format explained in the help file
%   locates section stringA in first step
%   locates keyword stringB in final step
%----------------------------------------------------------
{
    variable
        found,
        ffile = "sue.hlp",
        hfile = Null_String;

    hfile = expand_jedlib_file (ffile);

    ifnot (strlen (hfile))
    {
        flush ("File missing in library directory");
        return;
    }

    %   open temp buffer for help, name begins with space
    variable suehelp = " /Help Temp Buffer/";
    delbuf(suehelp);

    if (MINIBUFFER_ACTIVE)
        pop2buf(suehelp);
    else
        sw2buf(suehelp);

    %   insert into it copy of help file
    insert_file (hfile);
    bob ();

    CASE_SEARCH=1;
    found = bol_fsearch(stringA+" [#]");

    ifnot (found)
    {
        delbuf(suehelp);
        flush ("Not found: " + stringA);
        return;
    }

    eol();
    found = fsearch(stringB);

    ifnot (found)
    {
        delbuf(suehelp);
        stringB = substr (stringB, 1, strlen(stringB)-4);
        flush ("Not found: " + stringB);
        return;
    }

    %   both stringA and stringB found, cursor is on stringB

    eol();
    if(bfind("[#]")) del_to_eol();

    go_up(1);
    push_mark();
    bob();
    del_region();

    %   text above keyword deleted
    %   search for delimiter downstream

    ifnot (bol_fsearch("[#]"))
    {
        flush ("Delimiter not found, help file is damaged.");
        %   unlikely to happen

        delbuf(whatbuf);
        return;
    }

    %   delimiter found, delete text below it
    push_mark();
    eob();
    del_region();

    %   display help text at line 2
    if(bol_bsearch(stringB));
    else bob();
    while (is_whiteline) go_down(1);
    recenter(2);

    set_buffer_modified_flag (0);
    set_readonly(1);
    set_status_line ("  Press any key to exit", 0);

    dummy_menu(stringC);

    update(1);
    flush_input(getkey ());
    delbuf(suehelp);
}


define display_back2menu ()
{
    delbuf(whatbuf);
    menu_last_entry();
}


define display_newcomers ()
%----------------------------------------------------------
%   One screen orientation for newcomers
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable stringA = "PULLDOWN HELP";
    variable stringB = "    cursor  ";
    variable stringC = "&Quick Orientation for Newcomers";

    display_help (stringA,stringB,stringC);
    %   back to menu required
    menu_last_entry ();
}

define display_cmd ()
%----------------------------------------------------------
%   Pressing a key will reveal the corresponding function
%----------------------------------------------------------
{
    variable f, type, tmp, modif;

    flush ("Press key, command bound to it will show...");

    (type, f) = get_key_binding ();

    variable ks = Null_String;

    if (f == NULL)
    {
        flush ("Key undefined.");
        return;
    }

    %   type -1 same as f NULL
    %   type is now 10*modifier plus traditional values
    %   as used in the switch below
    %   separate modifier and actual type

    tmp = type;

    type = type mod 10;
    %   actual type

    modif = (tmp-type)/10;
    %   full modifier

    ks = keyname (LASTKEY,modif);

    %   ks may be empty

    ifnot(strlen(ks))    ks = LASTKEY;
    else if (modif==8) ks = "Alt "+ ks;
    else if (modif==5) ks = "ShiftCtrl " + ks;
    else if (modif==4) ks = "Ctrl "+ ks;
    else if (modif==1) ks = "Shift "+ ks;
    else if (modif)
    {
        %   should not happen
        flush("Unknown error.");
        return;
    }

    switch (type)
        {
            case 0:
            if (1 == is_defined (f))
                vmessage
                ("Key \"%s\" runs the intrinsic function \"%s\".", ks, f);
            else
                vmessage
                ("Key \"%s\" runs the SLang function \"%s\".", ks, f);
        }

        {
            case 1:
            vmessage
            ("Key \"%s\" runs the internal function \"%s\".", ks, f);
        }

        {
            case 2:
            vmessage
            ("Key \"%s\" runs the keyboard macro \"%s\".", ks, f);
        }

        {
            case 3:
            vmessage ("Key \"%s\" inserts \"%s\".", ks, f);
        }
    write_scratch (ks + "  " + f);
}


define display_doclist ()
%----------------------------------------------------------
%   Will list the docs directory.
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    %   list the docs directory
    variable dir = dircat (JED_ROOT, "doc/");
    disk_dirlist (dir);
}


define display_aboutsue ()
%----------------------------------------------------------
%   Will display details of jed, sue, slang as used here
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable found, ffile = "sue.hlp", hfile;
    variable stringA = "AboutSue";
    variable stringB = "Copyright";

    hfile = expand_jedlib_file (ffile);

    ifnot (strlen (hfile))
    {
        flush ("File missing in library directory");
        return;
    }

    %   open temp buffer for help
    variable suehelp = " /Help Temp Buffer/";
    sw2buf(suehelp);
    erase_buffer ();

    %   insert into it copy of help file
    insert_file (hfile);
    bob ();

    CASE_SEARCH=1;
    found = bol_fsearch(stringA+" [#]");

    ifnot (found)
    {
        delbuf(ffile);
        flush (stringA + "  not found.");
        return;
    }

    eol();
    found = fsearch(stringB);

    ifnot (found)
    {
        delbuf(ffile);
        flush (stringB + " not found.");
        return;
    }

    recenter(2);
    bol ();
    newline ();
    newline ();
    go_up(1);
    goto_column(5);

    vinsert ("Version:        %s", _jed_version_string);
    newline ();
    goto_column(5);

    vinsert ("SLang Version:  %s", _slang_version_string);

    newline ();

    bob();
    () = fsearch ("copyright");

    set_readonly (1);
    set_status_line ("  Press any key to exit", 0);
    update(1);

    flush_input(getkey ());
    delbuf(whatbuf);
}


define display_clean_manpage ()
%----------------------------------------------------------
%   Simplifies the manpage format so the manpage can
%   be shown in the editor.
%----------------------------------------------------------
{
    bob();

    %   remove _^H underscores
    replace ("_\010", Null_String);

    %   remove ovrstrike
    while (fsearch ("\010"))
    {
        del();
        del();
    }

    bob();              %12345678
    while (bol_fsearch ("        "))
        loop(4) del ();

    bob();              %1234567
    while (bol_fsearch ("       "))
        loop(3) del ();

    %   UTF-8 junk, jed problem, hopeless workaround
    replace ("","");
    replace ("","");

    flush (" ");
}


define display_manpage ()
%----------------------------------------------------------
%   Retrieves a man page and displays it in the editor
%   simplifying the format.
%----------------------------------------------------------
{
    current_mini_cmd = _function_name;

    variable
        cmdstr  = Null_String,
        subj    = Null_String,
        buf     = "manpage ",
        rsc     = 0,
        frombuf = whatbuf ();

    subj = read__mini ("man", Null_String, Null_String, 2);
    ifnot (strlen (subj)) return;

    buf = buf + subj;

    sw2buf (buf);
    erase_buffer ();
    set_buffer_no_autosave ();
    set_status_line (" %b %p %O F1=Help", 0);

    cmdstr = strcat
    ("man -D --no-hyphenation %s 2> ~/", string(random(-1,0)), ".tmp");

    rsc = run_shell_cmd (sprintf(cmdstr,subj));

    %     exit status
    % 0   Successful program execution.
    % 1   Usage, syntax or configuration file error.
    % 2   Operational error.
    % 3   A child process returned a non-zero exit status.
    % 16  At least one of the pages/files/keywords wasn’t found

    if (rsc>0 and rsc<16)
    {
        delbuf (buf);
        sw2buf (frombuf);
        reset_screen ();
        flush ("Invalid command. (Check man is 2.5.4 or later.)");
        return;
    }

    else
    {
        display_clean_manpage ();

        bob ();
        if (bobp==eobp)
        {
            delbuf (buf);
            sw2buf (frombuf);
            reset_screen ();
            flush ("Nothing found");
            return;
        }

        set_buffer_modified_flag (0);
    }
}


define display_info_doc ()
%----------------------------------------------------------
%   Will display an 'info' document in a buffer
%   with normal cursor movements, no node jumping
%----------------------------------------------------------
{
    current_mini_cmd = _function_name;

    variable
        subj    = Null_String,
        thisbf  = Null_String,
        tmpbf   = Null_String;

    subj = read__mini ("info", Null_String, Null_String, 2);
    ifnot (strlen (subj)) return;

    thisbf = whatbuf ();
    tmpbf = " info " + subj;

    delbuf(tmpbf);
    sw2buf(tmpbf);
    hide_status_indicators(1);
    set_status_line (" %b %p %O F1=Help", 0);

    () = run_shell_cmd(tmpbf);

    ifnot (MINIBUFFER_ACTIVE)
    {
        reset_screen();
        flush(" ");
    }
    %   essential since 'info' is very uncooperative

    bob();
    if(is_whiteline)
    {
        sw2buf(thisbf);
        delbuf(tmpbf);
        flush("Not found.");
        return;
    }
    newline();

    bol_fsearch("* Index::");
    newline();
    push_mark();
    () = fsearch("*******");
    go_up(2);
    del_region ();

    while(bol_fsearch("* Menu:"))
    {
        push_mark();
        go_down(2);
        bol_fsearch("\n");
        del_region();
    }

    bob();
    while(bol_fsearch("File: ")) delete_line();

    eob();
    if(bol_bsearch("*****"))
    {
        if (bsearch("Index"))
        {
            bol();
            push_mark();
            eob();
            del_region();
        }
    }

    bob();
    if (bol_fsearch("* Menu:"))
    {
        del_to_eol();
        insert("Chapters:");
        bob();
    }

    bob();              %1234567
    while (bol_fsearch ("       "))
        loop(4) del ();
    bob();
}


define display_calendar ()
%----------------------------------------------------------
%   Will display a calendar at the bottom of scratch buffer,
%   year is prompted for.
%----------------------------------------------------------
{
    current_mini_cmd = _function_name;

    variable
        s     = Null_String,
        yy    = strtok (time()," ")[4];

    s = strtrim
    (read__mini ("Calendar for year (4 digits): ", yy, yy, 2));

    if(strlen(s) != 4) return;

    if(string(_slang_guess_type(s)) == "Integer_Type")
        yy = s;

    %   if type not integer, default value for yy applies

    %   switch to scratch buffer to display calendar
    sw2buf("*scratch*");
    eob();
    newline();
    push_spot();

    () = run_shell_cmd ("cal -m " + yy);
    %   will insert output into current buffer

    pop_spot();
    recenter(2);
}


define display_ascii_table ()
%----------------------------------------------------------
%   Will display an ascii table at the bottom of scratch buffer
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable
        found,
        ffile = "sue.hlp",
        hfile;

    hfile = expand_jedlib_file (ffile);

    ifnot (strlen (hfile))
    {
        flush ("File missing in library directory");
        return;
    }

    variable tmpbf = whatbuf();

    %   open temp buffer for help
    variable suehelp = " /Help Temp Buffer/";
    delbuf(suehelp);
    sw2buf(suehelp);

    %   insert into it copy of help file
    insert_file (hfile);
    bob ();

    found = fsearch("Oct   Dec   Hex   Char");

    ifnot (found)
    {
        delbuf(suehelp);
        flush (ffile + " is corrupt.");
        return;
    }

    bol ();
    toggle_block(2);

    found = fsearch("077   63    3F");

    ifnot (found)
    {
        delbuf(suehelp);
        flush (ffile + " is corrupt.");
        return;
    }

    variable txt = bufsubstr ();

    delbuf(whatbuf);

    %   insert table in scratch buffer
    sw2buf("*scratch*");
    eob ();
    newline ();
    push_spot ();
    insert(txt);
    pop_spot ();

    sw2buf(tmpbf);
    flush("ASCII table appended to scratch buffer.");
}

define convert_base ()
{
    current_mini_cmd = _function_name;

    variable strnumber = Null_String;

    strnumber =
    read__mini ("Integer to convert:", Null_String, Null_String, 2);

    strnumber = strtrim (strnumber);
    ifnot(strlen(strnumber)) return;
    strnumber = strlow (strnumber);

    variable k;
    try
    {
        k = integer (strnumber);
    }
    catch AnyError:
    {
        flush("Invalid input.");
        sleep(2);
        flush_input();

        variable
            stringA = "MiniBuffer",
            stringB = current_mini_cmd,
            stringC = "&MiniBuffer Current Command";

        variable tmpbf = whatbuf;

        display_help (stringA,stringB,stringC);

        sw2buf(tmpbf);

        return;
    }

    variable strbase   = Null_String;
    variable strprmpt  = "Convert to base (B/O/D/X):";

    strbase = read__mini (strprmpt, Null_String, Null_String, 2);

    strbase = strtrim (strbase);

    ifnot(strlen(strbase) == 1)
    {
        flush("Invalid input.");
        return;
    }

    strbase = strlow (strbase);

    %   B not b
    if(strbase == "b") strbase = "B";
    else if(strbase == "o");
    else if(strbase == "d");
    else if(strbase == "x");
    else
    {
        flush("Invalid input.");
        return;
    }

    variable strstart1 = Null_String;
    strstart1 = substr (strnumber, 1, 1);

    variable strstart2 = Null_String;
    strstart2 = substr (strnumber, 1, 2);

    variable nada = "Nothing to do.";

    if (strstart2 == "0b")
    {
        if(strbase == "B")
        {
            flush(nada);
            return;
        }
    }

    else if (strstart2 == "0x")
    {
        if(strbase == "x")
        {
            flush(nada);
            return;
        }
    }

    else if(strstart1 == "0")
    {
        if(strbase == "o")
        {
            flush(nada);
            return;
        }
    }

    else
    {
        if(strbase == "d")
        {
            flush(nada);
            return;
        }
    }

    variable strfrmt = "%" + strbase;
    variable rsult   = Null_String;
    rsult = sprintf (strfrmt, k);

    write_scratch(rsult);
    flush(strbase + ":  " + rsult + " (also appended to scratch buffer.");
}


define guide_start ()
%----------------------------------------------------------
%   Brief guide for newcomers
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable ffile = "briefguide.hlp", hfile;

    hfile = expand_jedlib_file (ffile);

    ifnot (strlen (hfile))
    {
        flush ("File missing in library directory");
        return;
    }

    %   open temp buffer for help
    variable sueguide = "/brief guide/";
    delbuf (sueguide);
    sw2buf (sueguide);
    set_buffer_no_autosave ();

    %   insert into it copy of help file
    insert_file (hfile);

    set_buffer_modified_flag (0);
    set_readonly (1);
    set_status_line
    ("   %l/%L   |   Escape: exit   |   Space: jump   |   F1: help", 0);
    hide_status_indicators(1);

    guide_keymap ();
    bob();
    () = bol_fsearch ("1.");
    variable m = "&Brief Guide";
    dummy_menu(m);
}


define guide_goto ()
{
    %   does line begin with digit and dot?
    bol ();
    push_spot();
    if (3 != re_fsearch ("^[1-9]?[\.]?"))
    {
        flush("Shortcut works only on headings!");
        return;
    }

    pop_spot();
    push_mark();
    eol();
    variable sss = bufsubstr ();

    if(fsearch (sss)) return;
    bol();
    if(bsearch (sss)) return;

    flush("No match found: file possibly tampered with.");
}


define guide_help ()
%----------------------------------------------------------
%   ancillary function to guide_start ()
%----------------------------------------------------------
{
    variable
        stringA="Brief Guide",
        stringB="    [#]",
        stringC="&Quick Help: Brief Guide";

    display_help (stringA,stringB,stringC);
}

define guide_keymap ()
{
    reset_slim_keymap ();
    $1 = "guide_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<256; i++) undefinekey (char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left, $1);
    undefinekey (Key_Enter, $1);
    undefinekey (Key_Del, $1);

    definekey ("prghelp_clip",         "^C",        $1);
    definekey ("search_forward",        "^F",       $1);
    definekey ("slim_quit",             "^W",       $1);
    definekey ("next_line_cmd",         Key_Down,   $1);
    definekey ("eob",                   Key_End,    $1);
    definekey ("guide_goto",            Key_Space,  $1);
    definekey ("slim_quit",             Key_Escape, $1);
    definekey ("guide_help",            Key_F1,     $1);
    definekey ("search_again_forward",  Key_F3,     $1);
    definekey ("bob",                   Key_Home,   $1);
    definekey ("page_down",             Key_PgDn,   $1);
    definekey ("page_up",               Key_PgUp,   $1);
    definekey ("previous_line_cmd",     Key_Up,     $1);
}


define keyhelp_hlp ()
%----------------------------------------------------------
%   Quick help when navigating the Global, Ctrl, Alt
%   keymap dumpings
%----------------------------------------------------------
{
    variable
        stringA="KEYMAPS Global, Ctrl_Map, Alt_Map",
        stringB="    [#]",
        stringC="&Quick Help: Keybindings";

    display_help (stringA,stringB,stringC);
}


define keyhelp_cursor ()
%----------------------------------------------------------
%   Ancillary function: determines position of cursor
%   in groups plain, ctrl, alt of Global keymap
%----------------------------------------------------------
{
    variable r = 0;

    r = bol_fsearch ("Ctrl Keys");

    if (r) return 1;

    r = bol_fsearch ("Alt Keys");

    if (r) return 2;

    return 3;
}


define keyhelp_initstrings ()
%----------------------------------------------------------
%   Maps function names to explanatory text but only for
%   functions with a key shortcut. Function names are
%   in alphabetical order. Keep order when editing.
%   Keep functions names under 25 chars, comments to fit line.
%   When removing or adding, adjust array dimensions!
%   Removing: if editing forgotten, junk in array and
%             help list ingores it
%   Adding:   if editing forgotten, function name shows
%             instead of comment.
%----------------------------------------------------------
{
    explainkey [*,*] =
    [
     ["ask_close_buffer",         "close buffer safely"],
     ["block_all",                "select entire buffer"],
     ["block_append",             "copy to clipboard, append to scratch buffer"],
     ["block_cancel",             "cancel selection"],
     ["block_copy",               "copy to clipboard"],
     ["block_cut",                "cut to clipboard"],
     ["block_cut_append",         "cut to clipboard, append to scratch buffer"],
     ["block_fill_rectangle",     "fill a rectangle block"],
     ["block_paste",              "paste from the clipboard"],
     ["block_save",               "save selection to disk file"],
     ["block_toggle_lines",       "start or stop selection by whole lines"],
     ["block_toggle_rectangle",   "start or stop selection of rectangle"],
     ["bm_list",                  "jump to bookmark"],
     ["bm_next",                  "to next bookmark past cursor"],
     ["bm_prev",                  "to previous bookmark ahead of cursor"],
     ["bm_set",                   "set bookmark at current cursor position"],
     ["case_lower",               "turn selection lower case"],
     ["case_toggle_initial",      "reverse case of initial in current word"],
     ["case_upper",               "turn selection upper case"],
     ["del_line",                 "delete line incl. linebreak"],
     ["del_to_bol",               "delete left of cursor to begin of line"],
     ["del_to_eol",               "delete from cursor to end of line"],
     ["disk_new_file",            "open a new buffer with name of choice"],
     ["disk_open",                "open a file by name or display file list"],
     ["disk_save_buffer",         "save current buffer to disk file"],
     ["disk_save_preserve_mtime", "save buffer to file preserving time stamp"],
     ["disk_treeview",            "display disk tree, current directory as root"],
     ["display_doclist",          "display doc directory"],
     ["duplicate_ceiling_char",   "duplicate at cursor the character above it"],
     ["duplicate_floor_char",     "duplicate at cursor the character below it"],
     ["duplicate_line",           "duplicate current line"],
     ["ed_backtab",               "to previous tab stop; if selection present, unindent"],
     ["ed_delete_backward",       "delete character left of cursor"],
     ["ed_delete_here",           "delete character at cursor; if any selection, delete selection"],
     ["ed_delete_word",           "delete current word"],
     ["ed_evalbuffer",            "let interpreter check current buffer for syntax errors"],
     ["ed_quick_4untab",          "detab (untab) from cursor onward, 4 spaces per tab"],
     ["ed_quick_untab",           "detab (untab) from cursor onward with standard settings"],
     ["ed_softtab",               "insert spaces for tabs; if selection present, indent"],
     ["ed_undelete",              "undelete last deletion at the cursor"],
     ["ed_undo2",                 "undo last changes"],
     ["ed_whatchar",              "show coding details for character at cursor"],
     ["ed_wrap_and_jump",         "wrap this paragraph, jump to next one"],
     ["ed_wrap_lines",            "wrap to end of paragraph, stay on same line"],
     ["end_of_trail",             "to end of line, whitespace trail included"],
     ["execute_cmd",              "execute an internal command from the minibuffer"],
     ["exit_sue",                 "exit the editor, safely"],
     ["go_beginbuffer_select",    "to begin of buffer extending selection"],
     ["go_beginline_select",      "to begin of line extending selection"],
     ["go_begin_buffer",          "to begin of buffer"],
     ["go_begin_line",            "to begin of line"],
     ["go_bof",                   "cursor to bottom of current function"],
     ["go_cursorleft_select",     "move one character left extending selection"],
     ["go_cursorright_select",    "move one character right extending selection"],
     ["go_cursor_left",           "move one character left"],
     ["go_cursor_right",          "move one character right"],
     ["go_down_select",           "move down one line vertically extending selection"],
     ["go_down_vertically",       "move down one line, same column"],
     ["go_endbuffer_select",      "to end of buffer extending selection"],
     ["go_endline_select",        "to end of line extending selection"],
     ["go_end_buffer",            "to end of buffer"],
     ["go_end_line",              "to end of line"],
     ["go_pagedn_select",         "one page down extending selection"],
     ["go_pageup_select",         "one page up extending selection"],
     ["go_page_down",             "one page down"],
     ["go_page_up",               "one page up"],
     ["go_paragraph_down",        "to next paragraph"],
     ["go_paragraph_up",          "to previous paragraph"],
     ["go_tof",                   "cursor to top of current function"],
     ["go_to_file",               "go to file extracting path from current line"],
     ["go_to_matching_fence",     "go to the match for () [] {}"],
     ["go_up_select",             "move one line up vertically extending selection"],
     ["go_up_vertically",         "move one line up, same column"],
     ["go_wordleft_select",       "to previous word extending selection"],
     ["go_wordright_select",      "to next word extending selection"],
     ["go_word_left",             "to previous word"],
     ["go_word_right",            "to next word"],
     ["go__to",                   "pulldown the goto submenu to go somewhere"],
     ["insert_code_for_key",      "don't execute command keys, insert their keystring"],
     ["kbd_quit",                 "abort function currently executed"],
     ["macro_play",               "play (execute) a macro once"],
     ["macro_repeat",             "execute a macro querying how many times"],
     ["macro_start_or_stop",      "start or stop recording a macro"],
     ["menu_last_entry",          "back to the menus, last entry used"],
     ["newline",                  "insert new line (without indenting)"],
     ["newline_and_indent",       "insert a new line and possibly indent"],
     ["next_buffer",              "switch to next buffer"],
     ["next_buffer",              "switch to next buffer"],
     ["occur_functions",          "list C or SLang functions in current buffer"],
     ["otherwindow",              "move to other window in a split screen"],
     ["pcp",                      "cursor back to its previous position"],
     ["prev_buffer",              "switch to previous buffer"],
     ["prghelp_keyword",          "help on highlighted keyword"],
     ["redraw",                   "redraw screen"],
     ["repeat_command",           "repeat last command"],
     ["reset_screen",             "reset screen"],
     ["scroll_down",              "scroll down in place from the middle of screen"],
     ["scroll_up",                "scroll up in place from the middle of screen"],
     ["search_again_backward",    "repeat last search, backward"],
     ["search_again_forward",     "repeat last search, forward"],
     ["search_forward",           "simple search forward"],
     ["select_menubar",           "to the menubar"],
     ["select_pulldown_edit",     "pull down the edit menu"],
     ["select_pulldown_file",     "pull down the file menu"],
     ["select_pulldown_help",     "pull down the help menu"],
     ["select_pulldown_misc",     "pull down the misc menu"],
     ["select_pulldown_search",   "pull down the search menu"],
     ["select_pulldown_view",     "pull down the view menu"],
     ["set_wrap_column",          "set the buffer specific wrap column"],
     ["shell",                    "to scratch buffer; if already there, issue shell command"],
     ["s_srch",                   "search dialog for options and string"],
     ["toggle_windows",           "toggle between split and full screen"],
     ["toggle_overwrite",         "toggle between Insert and Overwrite mode"],
     ["toggle_wrap",              "start or stop wrapping text"],
     ["ttest",                    "test function for code under construction"],
     ["t_sort",                   "dialog to set sort options and start sort"]
    ];
}


define keyhelp_init_save ()
%----------------------------------------------------------
%   special buffer is saved prompting for a file name
%----------------------------------------------------------
{
    push_spot();
    variable tmpbf = whatbuf ();
    disk_save_as ();
    rename_buffer (tmpbf);
    sw2buf(tmpbf);
    keyhelp__keymap ();
    keyhelp_initstrings();
    pop_spot();
}


define prghelp__find ()
%----------------------------------------------------------
%   Ancillary function to Programming help
%----------------------------------------------------------
{
    current_mini_cmd = Null_String;

    variable sstr=Null_String;
    sstr = read__mini ("Search for:", Null_String, Null_String, 0);
    ifnot (strlen (sstr)) return;
    LAST_SEARCH=sstr;
    ifnot(fsearch(sstr)) flush("Not found.");
}


define prghelp_find_again ()
%----------------------------------------------------------
%   Ancillary function to Programming help
%----------------------------------------------------------
{
    ifnot (right(1)) return;
    ifnot(fsearch(LAST_SEARCH)) flush("Not found.");
}


define keyhelp__keymap ()
%----------------------------------------------------------
%   Initialize list for the Global, Ctrl, Alt keymap
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "keyhelp_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<256; i++) undefinekey (char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left, $1);
    undefinekey (Key_Enter, $1);
    undefinekey (Key_Escape, $1);

    definekey ("prghelp__find",         "^F",           $1);
    definekey ("keyhelp_init_save",     "^S",           $1);
    definekey ("display_back2menu",     "^W",           $1);
    definekey ("next_line_cmd",         Key_Down,       $1);
    definekey ("eob",                   Key_End,        $1);
    definekey ("display_back2menu",     Key_Escape,     $1);
    definekey ("keyhelp_hlp",           Key_F1,         $1);
    definekey ("prghelp_find_again",    Key_F3,         $1);
    definekey ("bob",                   Key_Home,       $1);
    definekey ("page_down",             Key_PgDn,       $1);
    definekey ("page_up",               Key_PgUp,       $1);
    definekey ("previous_line_cmd",     Key_Up,         $1);
}


define keyhelp_fnames ()
{
    variable i=0, L=0;

    L=length(explainkey[*,0]);

    ifnot (L) return;

    for (i=0;i<L;i++)
    {
        bob();
        while (re_fsearch(explainkey[i,0]+"$"))
            replace_chars (strlen(explainkey[i,0]),explainkey[i,1]);
    }
}

define keyhelp_tmp ()
{
    variable i=0, L=0;

    L=length(explainkey[*,0]);

    ifnot (L) return;

    for (i=0;i<L;i++)
    {
        bob();
        while (re_fsearch(explainkey[i,0]+"$"))
        {
            eol();
            insert("    ");
            insert(explainkey[i,1]);
        }
    }
}


define keyhelp_bindings ()
%----------------------------------------------------------
%   Some formatting for the keymap listings
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    delbuf (" /KeyBindings/");
    sw2buf (" /KeyBindings/");

    newline();
    insert ("[Plain Keys]");
    newline();
    newline();
    dump_bindings ("global");
    bob();
    while(fsearch("ESC TAB")) delete_line();

    eob();
    newline();
    newline();
    insert ("[Ctrl Keys]");
    newline();
    newline();
    dump_bindings ("ctrl_kmap");
    eob ();

    newline();
    newline();
    insert ("[Alt Keys]");
    newline();
    newline();
    dump_bindings ("alt_kmap");
    eob ();

    newline();
    newline();
    insert ("[Shift Keys]");
    newline();
    newline();
    dump_bindings ("shift_kmap");
    eob ();

    newline();
    newline();
    insert ("[ShiftCtrl Keys]");
    newline();
    newline();
    dump_bindings ("shiftctrl_kmap");

    %   suppress lines containing self-insert characters
    bob ();
    while (fsearch ("self_insert_cmd")) delete_line ();

    readable_keys ();

    %   sort plain keys, case-insensitive
    bob ();
    bol_fsearch ("^");
    %   the first keybindings are chars 1-31
    %   represented by ^ with a letter
    push_mark ();
    while (not is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    %   sort Ctrl keys, case-insensitive
    bob ();
    () = fsearch("Ctrl Keys");
    go_down(1);
    go_down(1);
    push_mark ();
    while (not is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    %   sort Alt keys, case-insensitive
    bob();
    () = fsearch("Alt Keys");
    go_down(1);
    go_down(1);
    push_mark ();
    while (not is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    %   KP_Enter case
    bob();
    () = fsearch("Alt Keys");
    () = fsearch("Enter");
    loop(8) del();
    insert("KP_Enter");

    %   sort Shift keys, case-insensitive
    bob();
    () = fsearch("Shift Keys");
    go_down(1);
    go_down(1);
    push_mark ();
    while (not is_whiteline) go_down(1);
    go_up(1);
    sort2 ();

    %   sort ShiftCtrl keys, case-insensitive
    bob();
    () = fsearch("ShiftCtrl Keys");
    go_down(1);
    go_down(1);
    push_mark ();
    while (not is_whiteline) go_down(1);
    go_up(1);
    sort2 ();
    flush (" ");

    bob ();
    replace("(Null_String)","");

    keyhelp__keymap ();
    keyhelp_initstrings();

    keyhelp_fnames ();
    %keyhelp_tmp ();

    bob();
    () = fsearch("Plain");

    set_readonly(1);
    set_status_line
    ("  Escape: exit  |  F1: help",0);
    variable m = "&Keybindings Global";
    dummy_menu(m);
}


define keyhelp_quick ()
%----------------------------------------------------------
%   Quick help on Global keys
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable stringA="Global";
    variable stringB="FILE/BUFFER";
    variable stringC="&Quick Help: Global";

    display_help (stringA,stringB,stringC);

    %   back to menu required
    menu_last_entry();
}


define keyhelp_functions ()
%----------------------------------------------------------
%   Quick tabular help on function keys
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable stringA="F-Keys";
    variable stringB="FUNCTION KEYS";
    variable stringC="&Function Keys";

    display_help (stringA,stringB,stringC);

    %   back to menu required
    menu_last_entry();
}


define keyhelp_minibuffer ()
%----------------------------------------------------------
%   List of Minibuffer keymap
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable
        stringA = "MiniBuffer",
        stringB = "F1",
        stringC = "&MiniBuffer Keybindings";

    display_help (stringA,stringB,stringC);
}


define prghelp_clip ()
%----------------------------------------------------------
%   Ancillary function to Programming help
%----------------------------------------------------------
{
    push_spot ();
    bol ();
    toggle_block(2);
    eol ();
    block_copy ();
    pop_spot ();
    flush("Line copied.");
}


define prghelp_init_save ()
%----------------------------------------------------------
%   special buffer is saved prompting for a file name
%----------------------------------------------------------
{
    push_spot();
    variable tmpbf = whatbuf ();
    disk_save_as ();
    rename_buffer (tmpbf);
    sw2buf(tmpbf);
    prghelp__keymap ();
    pop_spot();
}


define prghelp_inidispl_save ()
%----------------------------------------------------------
%   special buffer is saved prompting for a file name
%----------------------------------------------------------
{
    push_spot();
    variable tmpbf = whatbuf ();
    disk_save_as ();
    rename_buffer (tmpbf);
    sw2buf(tmpbf);
    prghelp_keymap();
    pop_spot();
}


define prghelp_keymap ()
%----------------------------------------------------------
%   Initialize listings in keyword help
%   default ctrl-f1
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "prghelp_d_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<256; i++) undefinekey (char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);
    undefinekey (Key_Del,   $1);

    definekey ("prghelp_clip",          "^C",       $1);
    definekey ("prghelp_inidispl_save", "^S",       $1);
    definekey ("slim_quit",             "^W",       $1);
    definekey ("next_line_cmd",         Key_Down,   $1);
    definekey ("slim_quit",             Key_Escape, $1);
    definekey ("previous_line_cmd",     Key_Up,     $1);
}


define prghelp_dig (stringH,stringK)
%----------------------------------------------------------
%   Details of Functions which are not in functions.hlp
%----------------------------------------------------------
{
    variable
        b_tmpbf = "/prghelp_dig/",
        addrss  = Null_String,
        cmdstr  = Null_String,
        fmask   = strcat (JED_ROOT,"/lib/*sl");

    %   open temporary buffer
    delbuf(b_tmpbf);
    sw2buf(b_tmpbf);

    if(strlen(stringH) < 8)
        stringH = strcat(stringH, " ", char(40));
    %   add a round parenthesis when string is short,
    %   it may be a component in other function names

    cmdstr = strcat
    ("grep -ins ", stringH, fmask, " 2> ~/", string(random(-1,0)), ".tmp &");

    () = run_shell_cmd (cmdstr);
    %   output goes into current buffer

    sw2buf(b_tmpbf);

    bob();
    while (fsearch(";")) delete_line();
    bob();

    if (fsearch("define"))
    {
        push_mark();
        bol();
        addrss = bufsubstr ();
    }

    else
    {
        flush ("Not found: " + stringK);
        return;
    }

    variable
        kfile = Null_String,
        kline = Null_String,
        kl    = 0;

    kfile = extract_element(addrss,0,':');
    ifnot(strlen(kfile))
    {
        flush ("Not found.");
        return;
    }

    kline = extract_element(addrss,1,':');
    ifnot(strlen(kline))
    {
        flush ("Not found.");
        return;
    }
    kl    = integer (kline);

    if(kl < 1)
    {
        flush ("Not found.");
        return;
    }

    erase_buffer ();
    insert_file (kfile);
    bob();
    goto_line (kl);
    go_up(1);
    bol();
    newline();
    newline();
    go_up(2);
    insert("     " + kfile + ", line " + kline + "  (s. also scratch buffer)");
    recenter(2);
    go_up(1);
    bol();
    del_to_eol();
    go_down(1);

    variable m = "&Details of Functions";
    dummy_menu(m);
    write_scratch (kfile + "  " + kline);

    set_status_line ("  Press any key to exit", 0);
    set_readonly(1);
    update(1);
    flush_input(getkey());
    delbuf(whatbuf);
}


define prghelp_display (stringK)
%----------------------------------------------------------
%   Ancillary function to Programming help (in menu).
%   Will display documentation from file functions.hlp
%----------------------------------------------------------
{
    %   stringK must not contain any regex operators
    %   some heuristics has been applied on it

    variable
        found,
        string0 = stringK,
        SL = 0,
        hfile   = expand_jedlib_file ("functions.hlp");

    ifnot (strlen(hfile))
    {
        flush ("File missing in library directory");
        return;
    }

    %   most SLang functions begin with SL
    %   we may get an odd catch, rarely

    if (1 == is_substr (string0,"SL"))
    {
        stringK = "[0-9]\.  " + string0 + "$";
        SL = 1;
    }
    else
        stringK = "^" + string0 + "$";

    found = search_file (hfile, stringK, 1);

    ifnot (found)
    {
        %   further search in library directory

        variable stringH = Null_String;

        stringH =
        "'^d" + "efine " + string0 + "' ";
        %   line begins with "define" and function name
        %   waste of time if searching for variable

        prghelp_dig (stringH,string0);
        delbuf("/prghelp_dig/");

        if (bufferp("/prghelp_listings/"))
            sw2buf("/prghelp_listings/");

        return;
    }

    %   open temp buffer for help
    variable prghelp = " " + slash + string(_time) + slash;
    delbuf(prghelp);
    sw2buf(prghelp);
    set_buffer_no_autosave ();

    variable
        m = "&Details of Functions/Variables",
        delimL = Null_String;

    loop (50) delimL += "-";

    dummy_menu(m);

    if(SL)
    {
        () = insert_file (hfile);
        bob();
        () = re_fsearch(stringK);
        bol();
        push_mark();
        eol();
        string0 = bufsubstr();
        erase_buffer();
    }

    if(SL)
        found = insert_file_region (hfile, string0, delimL);
    else
        found = insert_file_region (hfile, string0+"\n", delimL);

    if(found < 2)
    {
        %   just in case, should not happen

        flush("Not found: " + string0);
        delbuf(whatbuf);
        return;
    }
    insert(delimL);
    bob();
    newline();

    update (1);
    set_readonly(1);
    prghelp_keymap ();

    set_status_line
    ("  Escape: exit   |   Up/Down: move   |   ctrl-c: copy line   |   ctrl-s: save",0);
}


define prghelp_enter ()
%----------------------------------------------------------
%   Details for functions listed in Programming help
%----------------------------------------------------------
{
    if (4 > what_line) return;

    %   no further details for sl-variables
    push_spot();
    bob();
    if (0==fsearch("Press Enter"))
    {
        pop_spot();
        return;
    }
    pop_spot();

    bol();
    if (what_char () != ' ') return;
    begin_of_word_right ();
    push_mark ();
    eol ();

    variable stringK=Null_String;
    stringK = strtrim(bufsubstr ());

    if (strlen(stringK)>1) prghelp_display (stringK);
    else flush("Keyword for search too short.");
}


define prghelp_list_hlp ()
%----------------------------------------------------------
%   Quick help when navigating lists from submenu Programming
%----------------------------------------------------------
{
    variable
        stringA="KEYMAPS Global, Ctrl_Map, Alt_Map",
        stringB="    [#]",
        stringC="&Quick Help: Intrinsics/Variables";

    display_help (stringA,stringB,stringC);
}


define prghelp__keymap ()
%----------------------------------------------------------
%   Initialize lists for submenu Programming help
%   like intrinsics, variable etc.
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "prghelp_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<256; i++) undefinekey (char(i), $1);

    undefinekey ("^S", $1);

    definekey ("prghelp__find",         "^F",       $1);
    definekey ("prghelp_init_save",     "^S",       $1);
    definekey ("display_back2menu",     "^W",       $1);
    definekey ("eob",                   Key_End,    $1);
    definekey ("prghelp_enter",         Key_Enter,  $1);
    definekey ("display_back2menu",     Key_Escape, $1);
    definekey ("prghelp_list_hlp",      Key_F1,     $1);
    definekey ("prghelp_find_again",    Key_F3,     $1);
    definekey ("bob",                   Key_Home,   $1);
    definekey ("page_down",             Key_PgDn,   $1);
    definekey ("page_up",               Key_PgUp,   $1);
}


define prghelp_listings (s)
%----------------------------------------------------------
%   Will list active functions/variables,
%   based on the global keymap and customization.
%   Returns nothing.
%----------------------------------------------------------
{
    variable a = _apropos ("Global", "", s);

    a = a[array_sort(a, "strcmpi")];

    variable strbf = "/prghelp_listings/";
    delbuf (strbf);
    sw2buf (strbf);
    set_buffer_no_autosave ();

    variable
        m    = Null_String,
        lin2 = Null_String,
        lin3 = Null_String;

    lin2 = "Press Enter on highlighted item to get details.";

    if (s==1)
        m = " &C-coded Intrinsics";

    else if (s==2)
    {
        m = " &SL-coded Functions";
        lin3 = "[ ancillary functions are not documented ]";
    }

    else if (s==4)
    {
        m = " &C-coded Variables";
        lin3 = "[ jed documentation applies only partially to sue ]";
    }

    else if (s==8)
    {
        m = " &SL-coded Variables";
        lin2 = "No further details, try and grep the Editor's library.";
    }

    else
    {
        delbuf(whatbuf);
        return;
    }

    dummy_menu(m);

    foreach (__tmp(a))
    {
        insert ("    ");
        insert(());
        newline ();
    }

    if (s==2)
    {
        bob();
        while (fsearch("dls_")) delete_line();
        bob();
        while (fsearch("tree_")) delete_line();
        bob();
        while (fsearch("viewer_")) delete_line();
        bob();
        while (fsearch("help_")) delete_line();
        bob();
        while (fsearch("recent_")) delete_line();
        bob();
        while (fsearch("pulldown_")) delete_line();
        bob();
        while (fsearch("display_")) delete_line();
        bob();
        while (fsearch("occur_")) delete_line();
        bob();
        while (fsearch("mini_")) delete_line();
        bob();
        while (fsearch("bm_")) delete_line();
    }

    bob ();
    newline();
    insert ("    ");
    insert (lin2);
    newline();

    if(strlen(lin3))
    {
        insert ("    ");
        insert(lin3);
        newline();
    }

    newline();
    bob();
    () = fsearch("    ");
    set_readonly(1);

    prghelp__keymap ();

    set_status_line
    ("  %p  |  Escape: exit  |  F1: Help", 0);
}


define prghelp_intrinsics ()
%----------------------------------------------------------
%   Lists C coded intrinsics in submenu Programming help
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable s=0x1;
    prghelp_listings (s);
}


define prghelp_slfunctions ()
%----------------------------------------------------------
%   Lists SL intrinsics in submenu Programming help
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable s=0x2;
    prghelp_listings (s);
}


define prghelp_c_vars ()
%----------------------------------------------------------
%   Lists C coded variables in submenu Programming help
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable s=0x4;
    prghelp_listings (s);
}


define prghelp_sl_vars ()
%----------------------------------------------------------
%   Lists SL coded variables in submenu Programming help
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable s=0x8;
    prghelp_listings (s);
}


define prghelp_keyword ()
%----------------------------------------------------------
%   can be called from normal editing buffer
%   to read details about keyword in documentation,
%   keyword is determined by selection or by current word
%----------------------------------------------------------
{
    variable
        fflag = 0,
        kword = Null_String;

    if (chk_stream==1)
    {
        %   if there is a selection on a single line
        %   use it as keyword

        () = dupmark ();
        kword = strtrim(bufsubstr());
    }

    else
    {
        %   is cursor on a word?
        if(is_posixwhite) return;

        %   there might be an unsuitable mark
        toggle_block (0);

        select_word();
        kword = bufsubstr ();
        begin_of_word_left();
    }

    %   any whitespace in selection?
    fflag  = is_substr(kword," ");
    fflag += is_substr(kword,"\t");

    %   any dot in selection?
    fflag += is_substr(kword,".");

    %   keyword too short?
    if (strlen(kword) < 2) fflag++;

    if (fflag)
    {
        flush ("String unsuitable for keyword search.");
        return;
    }

    %   we got the keyword
    prghelp_display (kword);
}


%**********************************************************
%   Functions to list and activate color schemes
%**********************************************************


define color_bodyline ()
{
    eol ();
    go_left(3);

    if(looking_at (".sl")) return 1;
    else return 0;

    %   this to check the line is a body line
}


define color_up ()
{
    if(color_bodyline)
    {
        go_up(1);
        if(is_whiteline) go_down (1);
        return;
    }

    if (what_line < 4)
        () = fsearch (".sl");
    else
        () = bsearch (".sl");

    bol();
}


define color_down ()
{
    if(color_bodyline)
    {
       go_down(1);
       if(is_whiteline) go_up (1);
       return;
    }

    if (what_line < 4)
        () = fsearch (".sl");
    else
        () = bsearch (".sl");

    bol();
}


define color_enter ()
%----------------------------------------------------------
%   Used in the color selection to activate a color scheme
%----------------------------------------------------------
{
    eol ();
    go_left(3);

    ifnot(looking_at (".sl")) return;
    %   this to check the line is a body line

    push_mark();
    bol();
    skip_white ();

    variable cscheme = bufsubstr ();

    cscheme = strcat (JED_ROOT, "/lib/colors/", cscheme);

    require(cscheme);

    delbuf(whatbuf);
}


define color_keymap ()
%----------------------------------------------------------
%   Defines the keys for the color selection screen.
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "color_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<256; i++) undefinekey (char(i), $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Del,   $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);

    definekey ("color_up",      Key_Up,     $1);
    definekey ("color_down",    Key_Down,   $1);
    definekey ("color_enter",   Key_Enter,  $1);
    definekey ("slim_quit",     "^W",       $1);
    definekey ("slim_quit",     Key_Escape, $1);
}


define color_start ()
%----------------------------------------------------------
%   Starts the color selection screen.
%----------------------------------------------------------
{
    if(nwinz) return;
    %   split screen

    variable colorbf = string(_time);
    sw2buf (" " + slash+colorbf+slash);
    erase_buffer ();

    variable
        Color_Scheme_Path = Null_String,
        lstD              = Null_String;

    Color_Scheme_Path = strcat (JED_ROOT, "/lib/colors/");
    lstD = listdir (Color_Scheme_Path);

    if (lstD == NULL) return;
    lstD = lstD[array_sort(lstD)];

    variable lenL = length(lstD);
    ifnot (lenL) return;

    variable i=0, l=0, p=0;

    for (i=0;i<lenL;i++)
    {
        l = strlen(lstD[i]);
        p = is_substr (lstD[i],".sl");

        if (p > 0 and l-p == 2)
        {
            insert ("    ");
            insert(lstD[i]);
            newline();
        }
    }

    if (bobp)
    {
        delbuf(whatbuf());
        flush ("No color schemes found.");
        return;
    }

    bob ();
    newline ();
    insert ("    ");
    insert("Press Enter on a color scheme to activate it.");
    newline();
    newline();
    () = bsearch(".sl");

    color_keymap ();
    set_readonly(1);
    set_status_line("  Escape: exit  |  Up, Down: navigate  |  Enter: select",0);
    variable m = "&Color Schemes";
    dummy_menu(m);
}


%**********************************************************

define select_pulldown_file ()
{
    menu_select_menu ("Global.&File");
}


define select_pulldown_edit ()
{
    menu_select_menu ("Global.&Edit");
}


define select_pulldown_search ()
{
    menu_select_menu ("Global.&Search");
}


define select_pulldown_view ()
{
    menu_select_menu ("Global.&View");
}

define select_pulldown_misc ()
{
    menu_select_menu ("Global.&Misc");
}


define select_pulldown_help ()
{
    menu_select_menu ("Global.&Help");
}


define suemenu ()
{
    variable m;

    m = "Global";
    menu_create_menu_bar (m);
    menu_use_menu_bar (m);

    menu_delete_items (m);
    menu_append_popup (m, "&File");
    menu_append_popup (m, "&Edit");
    menu_append_popup (m, "&Search");
    menu_append_popup (m, "&View");
    menu_append_popup (m, "&Misc");
    menu_append_popup (m, "&Help");

    m = "Global.&File";
    menu_delete_items (m);
    menu_append_item (m, "&New..",                  "disk_new_file");
    menu_append_item (m, "&Open..",                 "disk_open");
    menu_append_item (m, "&Close",                  "close_buffer");
    menu_append_item (m, "&Save",                   "disk_save_buffer");
    menu_append_item (m, "Save &As..",              "disk_save_as");
    menu_append_item (m, "&Insert..",               "disk_read_and_insert");
    menu_append_separator (m);
    menu_append_item (m, "Switch &Buffer..",        "bufed_start");
    menu_append_item (m, "Recent &Files..",         "recent_files");
    menu_append_item (m, "Recent &Dirs..",          "recent_dirs");
    menu_append_item (m, "&Tree..",                 "disk_treeview");
    menu_append_separator (m);
    menu_append_item (m, "&Properties",             "ed_properties");
    menu_append_separator (m);
    menu_append_item (m, "E&xit..",                 "exit_sue");
    menu_append_item (m, "&Quit",                   "quit_sue");

    m = "Global.&Edit";
    menu_delete_items (m);
    menu_append_popup (m,                           "&Block Operations");
    menu_append_popup (m,                           "&Line");
    menu_append_popup (m,                           "Ca&se");
    menu_append_popup (m,                           "&Insert");
    menu_append_popup (m,                           "&Trim");
    menu_append_popup (m,                           "&Wrap");
    menu_append_separator (m);
    menu_append_item  (m, "&Undo",                  "ed_undo2");
    menu_append_item  (m, "U&ntab..",               "ed_untab");

    m = "Global.&Edit.&Block Operations";
    menu_append_item (m, "Select &All",             "block_all");
    menu_append_item (m, "&Cancel Selection",       "block_cancel");
    menu_append_item (m, "&Rectangle on/off",       "block_toggle_rectangle");
    menu_append_item (m, "&Lines on/off",           "block_toggle_lines");
    menu_append_separator (m);
    menu_append_item (m, "Cop&y",                   "block_copy");
    menu_append_item (m, "Cu&t",                    "block_cut");
    menu_append_item (m, "&Paste",                  "block_paste");
    menu_append_item (m, "&Write to File..",        "block_save");
    menu_append_separator (m);
    menu_append_item (m, "&Indent",                 "block_shift_right");
    menu_append_item (m, "&Unindent",               "block_shift_left");
    menu_append_item (m, "&Fill Rectangle..",       "block_fill_rectangle");
    menu_append_item (m, "&Sort..",                 "t_sort");

    m = "Global.&Edit.&Line";
    menu_append_item (m, "D&uplicate",              "duplicate_line");
    menu_append_item (m, "&Delete",                 "del_line");
    menu_append_item (m, "Del to &EOL",             "del_to_eol");
    menu_append_item (m, "&Toggle EOL",             "toggle_crmode");

    m = "Global.&Edit.Ca&se";
    menu_delete_items (m);
    menu_append_item (m, "&Upper Case",             "case_upper");
    menu_append_item (m, "&Lower Case",             "case_lower");
    menu_append_item (m, "Toggle &Initial",         "case_toggle_initial");

    m = "Global.&Edit.&Insert";
    menu_delete_items (m);
    menu_append_item (m, "&Date, Time",             "ed_ins_datetime");
    menu_append_item (m, "&Buffer Name",            "insert(whatbuf)");
    menu_append_item (m, "&File Name",              "ed_ins_pathfile");
    menu_append_item (m, "&Literal",                "insert_code_for_key");

    m = "Global.&Edit.&Trim";
    menu_delete_items (m);
    menu_append_item (m, "Selectio&n",              "trim_lines");
    menu_append_item (m, "Around &Cursor",          "trim");
    menu_append_item (m, "&Vertical Trim",          "vertical_trim");
    menu_append_item (m, "Trim on &Save..",         "trimsave");

    m = "Global.&Edit.&Wrap";
    menu_delete_items (m);
    menu_append_item  (m, "Wrap Some &Lines",       "ed_wrap_lines");
    menu_append_item  (m, "Toggle Wrap &Mode",      "toggle_wrap");
    menu_append_item  (m, "Wrap &Column..",         "set_wrap_column");
    menu_append_item  (m, "&Unwrap",                "ed_unwrap");

    m = "Global.&Search";
    menu_delete_items (m);
    menu_append_popup (m,                           "&Search");
    menu_append_item  (m, "&Replace..",             "r_srch");
    menu_append_item  (m, "&Occurrences..",         "o_srch");
    menu_append_separator (m);
    menu_append_item (m,  "&List Functions",        "occur_functions");
    menu_append_item  (m, "Find in &Files..",       "ed_grep");
    menu_append_separator (m);
    menu_append_item  (m, "Set &Bookmark",          "bm_set");
    menu_append_popup (m,                           "&Goto");

    m = "Global.&Search.&Search";
    menu_delete_items (m);
    menu_append_item (m,  "&New Search..",          "s_srch");
    menu_append_item (m,  "Repeat &Forward",        "search_again_forward");
    menu_append_item (m,  "Repeat &Backward",       "search_again_backward");

    m = "Global.&Search.&Goto";
    menu_delete_items (m);
    menu_append_item (m,  "&Line..",                "go_to_line");
    menu_append_item (m,  "&Column..",              "go_to_column");
    menu_append_item (m,  "&Matching Fence",        "go_to_matching_fence");
    menu_append_item (m,  "&Previous Position",     "pcp");
    menu_append_item (m,  "&Bookmark..",            "bm_list");
    menu_append_item (m,  "Top of &Function",       "go_tof");
    menu_append_item (m,  "End of &Function",       "go_bof");

    m = "Global.&View";
    menu_delete_items (m);
    menu_append_item (m, "&Split on/off",           "toggle_windows");
    menu_append_item (m, "&Lockstep on/off",        "lock_step");
    menu_append_item (m, "&Other Window",           "otherwindow");
    menu_append_separator (m);
    menu_append_item (m, "Read-onl&y on/off",       "ed_manage_readonly");
    menu_append_item (m, "&Redraw",                 "redraw");
    menu_append_item (m, "SUE &Defaults",           "sue_defaults");
    menu_append_item (m, "&Color Schemes..",        "color_start");

    m = "Global.&Misc";
    menu_delete_items (m);
    menu_append_item (m, "&ASCII Table",            "display_ascii_table");
    menu_append_item (m, "Calen&dar..",             "display_calendar");
    menu_append_item (m, "&Calculator..",           "calc");
    menu_append_item (m, "Convert &Base..",         "convert_base");
    menu_append_separator (m);
    menu_append_item (m, "&Evaluate Buffer",        "ed_evalbuffer");
    menu_append_popup (m,                           "&Macros");
    menu_append_separator (m);
    menu_append_item (m, "&Which Command..",        "display_cmd");
    menu_append_item (m, "&Shell Command..",        "shell");
    menu_append_item (m, "&Internal Command..",     "execute_cmd");

    m = "Global.&Misc.&Macros";
    menu_delete_items (m);
    menu_append_item (m, "&Macro Start/Stop",       "macro_start_or_stop");
    menu_append_item (m, "Macro &Play",             "macro_play");
    menu_append_item (m, "Macro &Repeat..",         "macro_repeat");

    m = "Global.&Help";
    menu_delete_items (m);
    menu_append_item (m, "&Here's the Beef",        "display_newcomers");
    menu_append_item (m, "&User Guide",             "guide_start");
    menu_append_popup (m,                           "&Keys");
    menu_append_popup (m,                           "&Programming");
    menu_append_separator (m);
    menu_append_item (m, "Sundry &Docs..",          "display_doclist");
    menu_append_item (m, "&Man Pages..",            "display_manpage");
    menu_append_item (m, "&Info Docs..",            "display_info_doc");
    menu_append_separator (m);
    menu_append_item (m, "&About",                  "display_aboutsue");

    m = "Global.&Help.&Keys";
    menu_append_item (m, "&Quick",                  "keyhelp_quick");
    menu_append_item (m, "&General..",              "keyhelp_bindings");
    menu_append_item (m, "&Function Keys",          "keyhelp_functions");
    menu_append_item (m, "&Shiftkey Specials..",    "occur_shiftkeys");
    menu_append_item (m, "&Minibuffer ",            "keyhelp_minibuffer");

    m = "Global.&Help.&Programming";
    menu_append_item (m, "&Intrinsics..",           "prghelp_intrinsics");
    menu_append_item (m, "SL &Functions..",         "prghelp_slfunctions");
    menu_append_item (m, "&C Variables..",          "prghelp_c_vars");
    menu_append_item (m, "SL &Variables..",         "prghelp_sl_vars");
}



%**********************************************************
%   viewer functions
%**********************************************************

define viewer_pulldown_file ()
{
    menu_select_menu ("viewer.&File");
}

define viewer_pulldown_search ()
{
    menu_select_menu ("viewer.&Search");
}

define viewer_pulldown_help ()
{
    menu_select_menu ("viewer.&Help");
    %   it works in other keymaps so no need to bother
}

define viewer_readable_keys()
{
    bob();
    replace("viewer_exit" , "exit after confirming");

    bob();
    replace("viewer_paste" , "paste to minibuffer");

    bob();
    replace("prghelp_clip" , "copy current line to clipboard");

    bob();
    replace("block_save" , "save selection");

    bob();
    replace("block_toggle_lines" , "selection on/off (by lines)");

    bob();
    replace("bob" , "begin of file");

    bob();
    replace("eob" , "end of file");

    bob();
    replace("quit_sue" ,  "exit viewer, no questions asked");

    bob();
    replace("next_line_cmd" , "next line");

    bob();
    replace("previous_line_cmd" , "previous line");

    bob();
    replace("search_again_forward" , "search again forward");

    bob();
    replace("search_forward" , "search forward");

    bob();
    replace("select_menubar" , "menubar");

    bob();
    replace("viewer_left" , "move left on long lines");

    bob();
    replace("viewer_pulldown_file" , "pull down File menu");

    bob();
    replace("viewer_pulldown_help" , "pull down Help menu");

    bob();
    replace("viewer_pulldown_help" , "pull down Help menu");

    bob();
    replace("viewer_pulldown_search" , "pull down Search menu");

    bob();
    replace("viewer_right" , "move right on long lines");

    bob();
    replace("viewer_scroll_down" , "scroll down");

    bob();
    replace("viewer_scroll_up" , "scroll up");

    bob();
    replace("viewer_go_line" , "go to line");
}


define viewer_help ()
{
    variable thiskeymap = what_keymap();

    sw2buf (" /ViewerKeys/");

    newline();
    newline();

    dump_bindings (thiskeymap);

    readable_keys();
    viewer_readable_keys();

    %   sort keys, case-insensitive
    bob ();
    () = fsearch("Press");
    go_down(1);
    begin_of_word_right ();

    push_mark ();
    eob();
    while (bolp) go_left(1);
    %   cursor at end of last non-empty line

    sort2 ();
    eob();
    newline();
    newline();
    flush(" ");
    set_readonly(1);

    bob ();
    while(is_whiteline) go_down(1);

    set_status_line ("  Press any key to exit  ", 0);
    variable m = "&Keybindings Viewer";
    dummy_menu(m);
    update(1);
    flush_input(getkey);
    delbuf(whatbuf);
}

define viewer_go_line ()
{
    if (check_rectangle) return;
    current_mini_cmd = _function_name;

    shiftkey_check();

    variable s = 0;
    s = integer (read__mini ("Goto line:", Null_String, Null_String, 2));

    ifnot (s) return;

    goto_line (s);
    %   cursor is at bol
}


define viewer_occurrences ()
{
    occur_simple ();
}


define viewer_left()
{
    occur_left();
    %   moves sideways half a screen on line without cursor
}


define viewer_right ()
{
    occur_right();
    %   moves sideways half a screen on line without cursor
}


define viewer_scroll_down ()
%----------------------------------------------------------
%   Scroll down half screen on current screen line
%----------------------------------------------------------
{
    variable lm = window_info ('r') / 2;

    %   scroll up
    if(up (lm)) recenter (lm);
    bol();
}


define viewer_scroll_up ()
%----------------------------------------------------------
%   Scroll up half screen on current screen line
%----------------------------------------------------------
{
    variable lm = window_info ('r') / 2;

    %   scroll up
    if(down(lm)) recenter (lm);
    bol();
}


define viewer_exit()
{
    variable rply = Null_String;
    rply = strlow(read__mini("Exit Viewer (y/n)?", "y", "y", 2));
    if (rply == "y") exit_jed ();
}


define viewer_properties()
%----------------------------------------------------------
%   Will display buffer and file properties
%----------------------------------------------------------
{
    %   get file info before creating special buffer
    %   this intrinsic from userinfo.c pushes into the stack
    %   the name of owner and group for the current file
    get_owner_group();
    variable grouP = ();
    variable owneR = ();

    variable
        tmpbf    = Null_String,
        stringA  = " /on the fly/",
        stringB  = "&Properties",
        full_fn  = Null_String,
        thisfile = Null_String,
        thisdir  = Null_String,
        thisbf   = Null_String,
        fs_flag  = 0,
        fstatus  = 0,
        fmode    = 0,
        bf_flags = 0,
        bf_modif = Null_String,
        spaces5  = "     ",
        spaces4  = "    ";

    tmpbf = whatbuf ();
    fs_flag = fs_readonly ();
    fmode = is_readonly();

    (thisfile, thisdir, thisbf, bf_flags) = getbuf_info ();

    full_fn = thisdir + thisfile;
    fstatus = file_status (full_fn);

    %   open temp buffer
    delbuf(stringA);
    sw2buf(stringA);
    newline();
    newline();

    insert(spaces4);
    insert("Directory:");
    goto_column (21);
    insert(thisdir);
    newline();
    newline();

    insert(spaces4);
    insert("File on disk:");
    goto_column (21);
    if (1 == fstatus) insert(thisfile);

    if (fstatus==1)
    {
        variable st;
        st = stat_file (full_fn);

        newline();
        insert(spaces5);
        insert("permissions:");
        goto_column (21);
        %   convert mode to octal first
        insert(ed_permissions(sprintf("%o",st.st_mode)));

        newline();
        insert(spaces5);
        insert("owner:");
        goto_column (21);

        insert(string(owneR));

        newline();
        insert(spaces5);
        insert("group:");
        goto_column (21);
        insert(string(grouP));

        newline();
        insert(spaces5);
        insert("bytes:");
        goto_column (21);
        insert(string(st.st_size));

        newline();
        insert(spaces5);
        insert("modified:");
        goto_column (21);
        insert(strftime ("%Y-%m-%d %H:%M", localtime (st.st_mtime)));

        if (fs_flag > 0)
        {
            newline();
            insert(spaces5);
            insert("warning!");
            goto_column (21);
            insert("READ-ONLY file system");
        }
    }

    flush (strftime ("  %Y-%m-%d %H:%M", localtime (_time)));
    ed_display (stringA, stringB);

    sw2buf(tmpbf);
}

define viewer_unwrap ()
{
    current_mini_cmd = _function_name;

    EXIT_BLOCK set_readonly(1);
    set_readonly(0);

    variable ddir, ffile, lline, rr;
    bol();
    trim();
    push_mark();
    eol();
    trim();
    lline = bufsubstr();

    (ffile, ddir , , ) = getbuf_info ();
    erase_buffer();
    insert_file (ddir+ffile);
    bob();
    () = fsearch(lline);
}

define viewer_wrap ()
{
    current_mini_cmd = _function_name;

    EXIT_BLOCK set_readonly(1);

    set_readonly(0);
    WRAP = SCREEN_WIDTH - 2;

    ifnot(right(1)) newline();
    left(1);
    %   avoid complaint about stack underflow

    variable i=1;
    variable cc;
    push_spot();
    bob();

    forever
    {
        eol();
        trim();

        if(what_column > WRAP)
        {
            % look for a space
            goto_column(WRAP);

            for(i=0; i < 10; i++)
            {
                cc = what_char;

                if(cc != 32)
                {
                    go_left(1);
                    continue;
                }

                % space found
                break;
            }

            if(i==10) goto_column(WRAP);

            newline();
            % line has been split, cursor is down 1
        }
        ifnot(down(1)) break;
    }
    pop_spot();
}


define viewer_keymap ()
{
    reset_slim_keymap ();
    $1 = "viewer_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=0; i<256; i++) undefinekey (char(i), $1);

    definekey ("viewer_pulldown_file",      "f",        $1);
    definekey ("viewer_pulldown_search",    "s",        $1);
    definekey ("viewer_pulldown_help",      "h",        $1);
    definekey ("viewer_pulldown_help",      Key_F1,     $1);

    definekey ("select_menubar",            Key_F10,    $1);
    definekey ("previous_line_cmd",         Key_Up,     $1);
    definekey ("next_line_cmd",             Key_Down,   $1);
    definekey ("viewer_left",               Key_Left,   $1);
    definekey ("viewer_right",              Key_Right,  $1);
    definekey ("bob",                       Key_Home,   $1);
    definekey ("eob",                       Key_End,    $1);
    definekey ("viewer_scroll_up",          Key_PgDn,   $1);
    definekey ("viewer_scroll_down",        Key_PgUp,   $1);
    definekey ("block_toggle_lines",        Key_F12,    $1);
    definekey ("search_again_forward",      Key_F3,     $1);

    definekey ("prghelp_clip",              "^C",       $1);
    definekey ("search_forward",            "^F",       $1);
    definekey ("viewer_go_line",            "^G",       $1);
    definekey ("block_save",                "^S",       $1);
    definekey ("viewer_paste",              "^V",       $1);
    definekey ("viewer_exit",               Key_Escape, $1);
}


define viewer_init ()
{
    if (keymap_p("viewer_slimkmap"))
    {
        use_keymap("viewer_slimkmap");
        return;
    }

    %   if keymap does not exist, create it
    viewer_keymap ();
}


define viewer_paste ()
{
    if(MINIBUFFER_ACTIVE) call ("yank");

    else flush("Paste only to minibuffer!");
}



define viewer_hints ()
%----------------------------------------------------------
%   One screen info
%----------------------------------------------------------
{
    variable stringA = "Viewer";
    variable stringB = "    Specific Help";
    variable stringC = "&Quick Hints for Viewing";
    display_help (stringA,stringB,stringC);
}


define viewer_menu ()
{
    variable m;

    m = "viewer";
    menu_create_menu_bar (m);
    menu_use_menu_bar (m);

    menu_delete_items (m);
    menu_append_popup (m, "&File");
    menu_append_popup (m, "&Search");
    menu_append_popup (m, "&Help");

    m = "viewer.&File";
    menu_delete_items (m);
    menu_append_item (m, "P&roperties", "viewer_properties");
    menu_append_separator (m);
    menu_append_item (m, "Selection &on/off", "block_toggle_lines");
    menu_append_item (m, "&Save Selection", "block_save");
    menu_append_separator (m);
    menu_append_item (m, "&Copy", "prghelp_clip");
    menu_append_item (m, "&Paste", "viewer_paste");
    menu_append_separator (m);
    menu_append_item (m, "&Wrap",   "viewer_wrap");
    menu_append_item (m, "&Unwrap Lines", "viewer_unwrap");
    menu_append_separator (m);
    menu_append_item (m, "E&xit", "viewer_exit");

    m = "viewer.&Search";
    menu_delete_items (m);
    menu_append_item (m,  "&Search", "search_forward");
    menu_append_item (m,  "Search &Again", "search_again_forward");
    menu_append_item (m,  "&Occurrences", "viewer_occurrences");
    menu_append_separator (m);
    menu_append_item (m,  "&Go to Line", "viewer_go_line");

    m = "viewer.&Help";
    menu_delete_items (m);
    menu_append_item (m, "&Hints", "viewer_hints");
    menu_append_item (m, "&Keys", "viewer_help");
}


define viewer_load()
{
    if (is_substr(whatbuf,slash)) return;

    viewer_menu ();
    viewer_init ();

    %hide_status_indicators(1);
    set_status_line ("  %F  %l/%L  |  F1=Help", 1);
    flush(" ");
}

